from asyncio import Queue, Task
from typing import Any, Dict, List, Tuple, Optional, TypeVar

from aiohttp import ClientSession, TCPConnector

from resotocore.analytics import AnalyticsEventSender
from resotocore.cli.model import CLI
from resotocore.config import ConfigHandler, ConfigOverride
from resotocore.core_config import CoreConfig
from resotocore.db.db_access import DbAccess
from resotocore.graph_manager.graph_manager import GraphManager
from resotocore.infra_apps.package_manager import PackageManager
from resotocore.infra_apps.runtime import Runtime
from resotocore.message_bus import MessageBus
from resotocore.model.db_updater import GraphMerger
from resotocore.model.model_handler import ModelHandler
from resotocore.query.template_expander import TemplateExpander
from resotocore.report import Inspector
from resotocore.service import Service
from resotocore.task.subscribers import SubscriptionHandler
from resotocore.task.task_handler import TaskHandlerService
from resotocore.types import JsonElement
from resotocore.user import UserManagement
from resotocore.web.certificate_handler import CertificateHandler
from resotocore.worker_task_queue import WorkerTaskQueue

T = TypeVar("T")


class ServiceNames:
    config = "config"
    message_bus = "message_bus"
    event_sender = "event_sender"
    db_access = "db_access"
    model_handler = "model_handler"
    task_handler = "task_handler"
    worker_task_queue = "worker_task_queue"
    template_expander = "template_expander"
    forked_tasks = "forked_tasks"
    cli = "cli"
    config_handler = "config_handler"
    cert_handler = "cert_handler"
    inspector = "inspector"
    infra_apps_runtime = "infra_apps_runtime"
    infra_apps_package_manager = "infra_apps_package_manager"
    user_management = "user_management"
    graph_manager = "graph_manager"
    subscription_handler = "subscription_handler"
    graph_merger = "graph_merger"
    config_override = "config_override"
    http_session = "http_session"
    scheduler = "scheduler"
    core_config_handler = "core_config_handler"
    merge_outer_edges_handler = "merge_outer_edges_handler"
    event_emitter_periodic = "event_emitter_periodic"


class Dependencies(Service):
    def __init__(self, **deps: Any) -> None:
        self.lookup: Dict[str, Any] = deps

    def add(self, name: str, service: T) -> "T":
        self.lookup[name] = service
        return service

    def extend(self, **deps: Any) -> "Dependencies":
        self.lookup = {**self.lookup, **deps}
        return self

    @property
    def services(self) -> List[Service]:
        return [v for _, v in self.lookup.items() if isinstance(v, Service)]

    @property
    def config(self) -> CoreConfig:
        return self.lookup[ServiceNames.config]  # type: ignore

    @property
    def message_bus(self) -> MessageBus:
        return self.lookup[ServiceNames.message_bus]  # type:ignore

    @property
    def event_sender(self) -> AnalyticsEventSender:
        return self.lookup[ServiceNames.event_sender]  # type:ignore

    @property
    def db_access(self) -> DbAccess:
        return self.lookup[ServiceNames.db_access]  # type:ignore

    @property
    def model_handler(self) -> ModelHandler:
        return self.lookup[ServiceNames.model_handler]  # type:ignore

    @property
    def task_handler(self) -> TaskHandlerService:
        return self.lookup[ServiceNames.task_handler]  # type:ignore

    @property
    def worker_task_queue(self) -> WorkerTaskQueue:
        return self.lookup[ServiceNames.worker_task_queue]  # type:ignore

    @property
    def template_expander(self) -> TemplateExpander:
        return self.lookup[ServiceNames.template_expander]  # type:ignore

    @property
    def forked_tasks(self) -> Queue[Tuple[Task[JsonElement], str]]:
        return self.lookup[ServiceNames.forked_tasks]  # type:ignore

    @property
    def cli(self) -> CLI:
        return self.lookup[ServiceNames.cli]  # type:ignore

    @property
    def config_handler(self) -> ConfigHandler:
        return self.lookup[ServiceNames.config_handler]  # type:ignore

    @property
    def cert_handler(self) -> CertificateHandler:
        return self.lookup[ServiceNames.cert_handler]  # type:ignore

    @property
    def inspector(self) -> Inspector:
        return self.lookup[ServiceNames.inspector]  # type:ignore

    @property
    def infra_apps_runtime(self) -> Runtime:
        return self.lookup[ServiceNames.infra_apps_runtime]  # type:ignore

    @property
    def infra_apps_package_manager(self) -> PackageManager:
        return self.lookup[ServiceNames.infra_apps_package_manager]  # type:ignore

    @property
    def user_management(self) -> UserManagement:
        return self.lookup[ServiceNames.user_management]  # type:ignore

    @property
    def graph_manager(self) -> GraphManager:
        return self.lookup[ServiceNames.graph_manager]  # type:ignore

    @property
    def subscription_handler(self) -> SubscriptionHandler:
        return self.lookup[ServiceNames.subscription_handler]  # type:ignore

    @property
    def graph_merger(self) -> GraphMerger:
        return self.lookup[ServiceNames.graph_merger]  # type:ignore

    @property
    def config_override(self) -> ConfigOverride:
        return self.lookup[ServiceNames.config_override]  # type:ignore

    @property
    def http_session(self) -> ClientSession:
        session: Optional[ClientSession] = self.lookup.get(ServiceNames.http_session)
        if not session:
            connector = TCPConnector(limit=0, ssl=False, ttl_dns_cache=300)
            session = ClientSession(connector=connector)
            self.lookup[ServiceNames.http_session] = session
        return session

    async def stop(self) -> None:
        if ServiceNames.http_session in self.lookup:
            await self.http_session.close()



from resotocore.db.model import QueryModel
from resotocore.message_bus import MessageBus, Action
import logging
import asyncio
from asyncio import Task, Future
from typing import Optional, Tuple, List
from contextlib import suppress
from datetime import timedelta
from resotocore.model.graph_access import ByNodeId, NodeSelector
from resotocore.service import Service
from resotocore.task.model import Subscriber
from resotocore.ids import NodeId, SubscriberId
from resotocore.task.task_handler import TaskHandlerService
from resotocore.ids import TaskId
from resotocore.task.subscribers import SubscriptionHandler
from resotocore.db.db_access import DbAccess
from resotocore.model.model_handler import ModelHandler
from resotocore.query.query_parser import parse_query


log = logging.getLogger(__name__)

subscriber_id = SubscriberId("resotocore")
merge_outer_edges = "merge_outer_edges"


class MergeOuterEdgesHandler(Service):
    def __init__(
        self,
        message_bus: MessageBus,
        subscription_handler: SubscriptionHandler,
        task_handler_service: TaskHandlerService,
        db_access: DbAccess,
        model_handler: ModelHandler,
    ):
        self.message_bus = message_bus
        self.merge_outer_edges_listener: Optional[Task[None]] = None
        self.subscription_handler = subscription_handler
        self.subscriber: Optional[Subscriber] = None
        self.task_handler_service = task_handler_service
        self.db_access = db_access
        self.model_handler = model_handler

    async def merge_outer_edges(self, task_id: TaskId) -> Tuple[int, int]:
        pending_outer_edge_db = self.db_access.pending_deferred_edge_db
        pending_edges = await pending_outer_edge_db.get(task_id)
        if pending_edges:
            graph_db = self.db_access.get_graph_db(pending_edges.graph)
            model = await self.model_handler.load_model(pending_edges.graph)

            async def find_node_id(selector: NodeSelector) -> Optional[NodeId]:
                try:
                    if isinstance(selector, ByNodeId):
                        node = await graph_db.get_node(model, selector.value)
                        return node.get("id") if node else None
                    else:
                        query = parse_query(selector.query).with_limit(2)
                        async with await graph_db.search_list(QueryModel(query, model)) as cursor:
                            results = [node async for node in cursor]
                            if len(results) > 1:
                                log.warning(
                                    f"task_id: {task_id}: node selector {selector.query} returned more than one node."
                                    "The edge was not created."
                                )
                                return None

                        return next(iter(results), {}).get("id", None)  # type: ignore
                except Exception as e:
                    log.warning(f"task_id: {task_id}: Error {e} when finding node {selector}")
                    return None

            edges: List[Tuple[NodeId, NodeId, str]] = []
            for edge in pending_edges.edges:
                from_id = await find_node_id(edge.from_node)
                to_id = await find_node_id(edge.to_node)
                if from_id and to_id:
                    edges.append((from_id, to_id, edge.edge_type))

            updated, deleted = await graph_db.update_deferred_edges(edges, pending_edges.created_at)

            log.info(
                f"MergeOuterEdgesHandler: updated {updated}/{len(pending_edges.edges)},"
                f"  deleted {deleted} edges in task id {task_id}"
            )

            return (updated, deleted)
        else:
            log.info(f"MergeOuterEdgesHandler: no pending edges for task id {task_id} found.")

            return (0, 0)

    async def mark_done(self, task_id: TaskId) -> None:
        pending_outer_edge_db = self.db_access.pending_deferred_edge_db
        await pending_outer_edge_db.delete(task_id)

    async def __handle_events(self, subscription_done: Future[None]) -> None:
        async with self.message_bus.subscribe(subscriber_id, [merge_outer_edges]) as events:
            subscription_done.set_result(None)
            while True:
                event = await events.get()
                if isinstance(event, Action) and event.message_type == merge_outer_edges:
                    await self.merge_outer_edges(event.task_id)
                    await self.mark_done(event.task_id)
                    await self.task_handler_service.handle_action_done(event.done(subscriber_id))

    async def start(self) -> None:
        subscription_done = asyncio.get_event_loop().create_future()
        self.subscriber = await self.subscription_handler.add_subscription(
            subscriber_id, merge_outer_edges, True, timedelta(seconds=30)
        )
        self.merge_outer_edges_listener = asyncio.create_task(
            self.__handle_events(subscription_done), name=subscriber_id
        )
        await subscription_done

    async def stop(self) -> None:
        if self.merge_outer_edges_listener:
            with suppress(Exception):
                self.merge_outer_edges_listener.cancel()
        if self.subscriber:
            await self.subscription_handler.remove_subscription(subscriber_id, merge_outer_edges)



from abc import ABC, abstractmethod
from typing import Optional, List

from attr import define

from resotocore.task.task_description import Job, RunningTask, Workflow
from resotocore.ids import TaskDescriptorId


@define(frozen=True)
class RunningTaskInfo:
    running_task: RunningTask
    scheduled_next: bool = False


class TaskHandler(ABC):
    @abstractmethod
    async def list_workflows(self) -> List[Workflow]:
        pass

    @abstractmethod
    async def list_jobs(self) -> List[Job]:
        pass

    @abstractmethod
    async def add_job(self, job: Job, force: bool = False) -> None:
        """
        Add a job to the task handler. If force is True, the job name validations and other checks will be ignored.
        Used for system jobs that are added by resoto itself.

        Make sure to never use the force parameter if the job is added by the user.
        """

    @abstractmethod
    async def delete_job(self, job_id: str, force: bool = False) -> Optional[Job]:
        """
        Delete a job from the task handler. If force is True, the job name validations
        and other checks will be ignored.
        Used for system jobs that are added by resoto itself.

        Make sure to never use the force parameter if the job is added by the user.
        """

    @abstractmethod
    async def start_task_by_descriptor_id(self, uid: TaskDescriptorId) -> Optional[RunningTaskInfo]:
        pass

    @abstractmethod
    async def running_tasks(self) -> List[RunningTask]:
        pass



import warnings
from datetime import datetime
from typing import Callable, Any, List

from apscheduler.executors.asyncio import AsyncIOExecutor
from apscheduler.job import Job
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger

from resotocore.service import Service
from resotolib.utils import get_local_tzinfo


warnings.filterwarnings(
    "ignore",
    message="The localize method is no longer necessary, as this time zone supports the fold attribute",
)


class Scheduler(Service):
    def __init__(self) -> None:
        self.scheduler = AsyncIOScheduler(
            jobstores={"default": MemoryJobStore()},
            executors={"default": AsyncIOExecutor()},
            # coalesce: run once instead of many times if the job should be run more than once in succession
            # max_instances: allowed parallel instances for one job
            # misfire_grace_time: seconds after the designated runtime that the job is still allowed to be run
            job_defaults={"coalesce": True, "max_instances": 32, "misfire_grace_time": 3600},
            timezone=get_local_tzinfo(),
        )

    async def start(self) -> None:
        self.scheduler.start()

    async def stop(self) -> None:
        self.scheduler.remove_all_jobs()

    def cron(
        self, job_id: str, name: str, cron_string: str, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Job:
        trigger = CronTrigger.from_crontab(cron_string)
        return self.scheduler.add_job(func, trigger, args, kwargs, job_id, name)

    def at(self, job_id: str, name: str, dt: datetime, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Job:
        return self.scheduler.add_job(func, "date", args, kwargs, job_id, name, run_date=dt)

    def interval(
        self, job_id: str, name: str, every_second: int, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Job:
        return self.scheduler.add_job(func, "interval", args, kwargs, job_id, name, seconds=every_second)

    def list_jobs(self) -> List[Job]:
        return self.scheduler.get_jobs(jobstore="default")  # type: ignore



import asyncio
import logging
from typing import Optional, List, Dict, Tuple, Callable, AsyncIterator

from aiostream import stream
from attr import evolve, define

from resotocore.analytics import CoreEvent
from resotocore.cli.model import CLIContext, CLI
from resotocore.config import ConfigEntity, ConfigHandler
from resotocore.db.model import QueryModel
from resotocore.error import NotFoundError
from resotocore.ids import ConfigId, GraphName
from resotocore.model.model import Model
from resotocore.model.resolve_in_graph import NodePath
from resotocore.query.model import Aggregate, AggregateFunction, Query, P, AggregateVariable, AggregateVariableName
from resotocore.report import (
    Inspector,
    ReportCheck,
    Benchmark,
    BenchmarkResult,
    CheckCollection,
    CheckCollectionResult,
    CheckResult,
    CheckConfigPrefix,
    BenchmarkConfigPrefix,
    CheckConfigRoot,
    ResotoReportValues,
    BenchmarkConfigRoot,
    ResotoReportBenchmark,
    ResotoReportCheck,
    ReportSeverity,
    ReportSeverityPriority,
)
from resotocore.report.report_config import ReportCheckCollectionConfig, BenchmarkConfig
from resotocore.types import Json
from resotocore.util import value_in_path
from resotocore.service import Service

log = logging.getLogger(__name__)

CountByAccount = Dict[str, int]


def benchmark_id(name: str) -> ConfigId:
    return ConfigId(BenchmarkConfigPrefix + name)


def check_id(name: str) -> ConfigId:
    return ConfigId(CheckConfigPrefix + name)


@define
class CheckContext:
    accounts: Optional[List[str]] = None
    severity: Optional[ReportSeverity] = None
    only_failed: bool = False
    parallel_checks: int = 10

    def includes_severity(self, severity: ReportSeverity) -> bool:
        if self.severity is None:
            return True
        else:
            return ReportSeverityPriority[self.severity] <= ReportSeverityPriority[severity]


class InspectorService(Inspector, Service):
    def __init__(self, cli: CLI) -> None:
        self.config_handler: ConfigHandler = cli.dependencies.config_handler
        self.db_access = cli.dependencies.db_access
        self.cli = cli
        self.template_expander = cli.dependencies.template_expander
        self.model_handler = cli.dependencies.model_handler
        self.event_sender = cli.dependencies.event_sender

    async def start(self) -> None:
        # TODO: we need a migration path for checks added in existing configs
        config_ids = {i async for i in self.config_handler.list_config_ids()}
        overwrite = False  # only here to simplify development. True until we reach a stable version.
        # we renamed this config in 3.2.6 - old installations still might have it
        # this line can be removed in a future version
        await self.config_handler.delete_config(ConfigId("resoto.report.benchmark.aws_cis_1.5"))
        for name, js in BenchmarkConfig.from_files().items():
            if overwrite or benchmark_id(name) not in config_ids:
                cid = benchmark_id(name)
                log.info(f"Creating benchmark config {cid}")
                await self.config_handler.put_config(ConfigEntity(cid, {BenchmarkConfigRoot: js}), validate=False)
        for name, js in ReportCheckCollectionConfig.from_files().items():
            if overwrite or check_id(name) not in config_ids:
                cid = check_id(name)
                log.info(f"Creating check collection config {cid}")
                await self.config_handler.put_config(ConfigEntity(cid, {CheckConfigRoot: js}), validate=False)

    async def list_benchmarks(self) -> List[Benchmark]:
        return [
            await self.__benchmark(i)
            async for i in self.config_handler.list_config_ids()
            if i.startswith(BenchmarkConfigPrefix)
        ]

    async def benchmark(self, name: str) -> Optional[Benchmark]:
        try:
            return await self.__benchmark(benchmark_id(name))
        except ValueError:
            return None

    async def list_checks(
        self,
        *,
        provider: Optional[str] = None,
        service: Optional[str] = None,
        category: Optional[str] = None,
        kind: Optional[str] = None,
        check_ids: Optional[List[str]] = None,
    ) -> List[ReportCheck]:
        def inspection_matches(inspection: ReportCheck) -> bool:
            return (
                (provider is None or provider == inspection.provider)
                and (service is None or service == inspection.service)
                and (category is None or category in inspection.categories)
                and (kind is None or kind == inspection.result_kind)
                and (check_ids is None or inspection.id in check_ids)
            )

        return await self.filter_checks(inspection_matches)

    async def perform_benchmark(
        self,
        graph: GraphName,
        benchmark_name: str,
        *,
        accounts: Optional[List[str]] = None,
        severity: Optional[ReportSeverity] = None,
        only_failing: bool = False,
    ) -> BenchmarkResult:
        benchmark = await self.__benchmark(benchmark_id(benchmark_name))
        context = CheckContext(accounts=accounts, severity=severity, only_failed=only_failing)
        return await self.__perform_benchmark(benchmark, graph, context)

    async def perform_checks(
        self,
        graph: GraphName,
        *,
        provider: Optional[str] = None,
        service: Optional[str] = None,
        category: Optional[str] = None,
        kind: Optional[str] = None,
        check_ids: Optional[List[str]] = None,
        accounts: Optional[List[str]] = None,
        severity: Optional[ReportSeverity] = None,
        only_failing: bool = False,
    ) -> BenchmarkResult:
        checks = await self.list_checks(
            provider=provider, service=service, category=category, kind=kind, check_ids=check_ids
        )
        provider_name = f"{provider}_" if provider else ""
        service_name = f"{service}_" if service else ""
        category_name = f"{category}_" if category else ""
        kind_name = f"{kind}_" if kind else ""
        check_id_name = f"{check_ids[0]}_" if check_ids else ""
        title = f"{provider_name}{service_name}{category_name}{kind_name}{check_id_name}benchmark"
        benchmark = Benchmark(
            id=title,
            title=title,
            description="On demand benchmark",
            documentation="On demand benchmark",
            framework="resoto",
            version="1.0",
            checks=[c.id for c in checks],
            children=[],
        )
        context = CheckContext(accounts=accounts, severity=severity, only_failed=only_failing)
        return await self.__perform_benchmark(benchmark, graph, context)

    async def __benchmark(self, cfg_id: ConfigId) -> Benchmark:
        cfg = await self.config_handler.get_config(cfg_id)
        if cfg is None or BenchmarkConfigRoot not in cfg.config:
            raise ValueError(f"Unknown benchmark: {cfg_id}")
        return BenchmarkConfig.from_config(cfg)

    async def filter_checks(self, report_filter: Optional[Callable[[ReportCheck], bool]] = None) -> List[ReportCheck]:
        cfg_ids = [i async for i in self.config_handler.list_config_ids() if i.startswith(CheckConfigPrefix)]
        loaded = await asyncio.gather(*[self.config_handler.get_config(cfg_id) for cfg_id in cfg_ids])
        # fmt: off
        return [
            check
            for entry in loaded if isinstance(entry, ConfigEntity) and CheckConfigRoot in entry.config
            for check in ReportCheckCollectionConfig.from_config(entry) if report_filter is None or report_filter(check)
        ]
        # fmt: on

    async def list_failing_resources(
        self, graph: GraphName, check_uid: str, account_ids: Optional[List[str]] = None
    ) -> AsyncIterator[Json]:
        context = CheckContext(accounts=account_ids)
        return await self.__list_failing_resources(graph, check_uid, context)

    async def __list_failing_resources(
        self, graph: GraphName, check_uid: str, context: CheckContext
    ) -> AsyncIterator[Json]:
        checks = await self.list_checks(check_ids=[check_uid])
        if not checks:
            raise NotFoundError(f"Check {check_uid} not found")
        inspection = checks[0]
        # load model
        model = await self.model_handler.load_model(graph)
        # load configuration
        cfg_entity = await self.config_handler.get_config(ResotoReportValues)
        cfg = cfg_entity.config if cfg_entity else {}
        # final environment: defaults are coming from the check and are eventually overriden in the config
        env = inspection.environment(cfg)
        account_id_prop = "ancestors.account.reported.id"
        # if the result kind is an account, we need to use the id directly instead of walking the graph
        if (result_kind := model.get(inspection.result_kind)) and "account" in result_kind.kind_hierarchy():
            account_id_prop = "reported.id"

        async def perform_search(search: str) -> AsyncIterator[Json]:
            # parse query
            query = await self.template_expander.parse_query(search, on_section="reported", **env)
            # filter only relevant accounts if provided
            if context.accounts:
                query = Query.by(P.single(account_id_prop).is_in(context.accounts)).combine(query)
            async with await self.db_access.get_graph_db(graph).search_list(QueryModel(query, model)) as ctx:
                async for result in ctx:
                    yield result

        async def perform_cmd(cmd: str) -> AsyncIterator[Json]:
            # filter only relevant accounts if provided
            if context.accounts:
                account_list = ",".join(f'"{a}"' for a in context.accounts)
                cmd = f"search /{account_id_prop} in [{account_list}] | " + cmd
            cli_result = await self.cli.execute_cli_command(cmd, stream.list, CLIContext(env=env))
            for result in cli_result[0]:
                yield result

        if resoto_search := inspection.detect.get("resoto"):
            return perform_search(resoto_search)
        elif resoto_cmd := inspection.detect.get("resoto_cmd"):
            return perform_cmd(resoto_cmd)
        else:
            return stream.empty()  # type: ignore

    async def __perform_benchmark(
        self, benchmark: Benchmark, graph: GraphName, context: CheckContext
    ) -> BenchmarkResult:
        if context.accounts is None:
            context.accounts = await self.__list_accounts(benchmark, graph)

        perform_checks = await self.list_checks(check_ids=benchmark.nested_checks())
        check_by_id = {c.id: c for c in perform_checks if context.includes_severity(c.severity)}
        result = await self.__perform_checks(graph, perform_checks, context)
        await self.event_sender.core_event(CoreEvent.BenchmarkPerformed, {"benchmark": benchmark.id})

        def to_result(cc: CheckCollection) -> CheckCollectionResult:
            check_results = [
                CheckResult(check_by_id[cid], result.get(cid, {})) for cid in cc.checks or [] if cid in check_by_id
            ]
            children = [to_result(c) for c in cc.children or []]
            return CheckCollectionResult(
                cc.title, cc.description, documentation=cc.documentation, checks=check_results, children=children
            )

        top = to_result(benchmark).filter_result(context.only_failed)
        return BenchmarkResult(
            benchmark.title,
            benchmark.description,
            benchmark.framework,
            benchmark.version,
            documentation=benchmark.documentation,
            checks=top.checks,
            children=top.children,
            accounts=context.accounts,
            only_failed=context.only_failed,
            severity=context.severity,
        )

    async def __perform_checks(
        self, graph: GraphName, checks: List[ReportCheck], context: CheckContext
    ) -> Dict[str, CountByAccount]:
        # load model
        model = await self.model_handler.load_model(graph)
        # load configuration
        cfg_entity = await self.config_handler.get_config(ResotoReportValues)
        cfg = cfg_entity.config if cfg_entity else {}

        async def perform_single(check: ReportCheck) -> Tuple[str, CountByAccount]:
            return check.id, await self.__perform_check(graph, model, check, cfg, context)

        async with stream.map(
            stream.iterate(checks), perform_single, ordered=False, task_limit=context.parallel_checks
        ).stream() as streamer:
            return {key: value async for key, value in streamer}

    async def __perform_check(
        self, graph: GraphName, model: Model, inspection: ReportCheck, config: Json, context: CheckContext
    ) -> CountByAccount:
        # final environment: defaults are coming from the check and are eventually overriden in the config
        env = inspection.environment(config)
        account_id_prop = "ancestors.account.reported.id"
        # if the result kind is an account, we need to use the id directly instead of walking the graph
        if (result_kind := model.get(inspection.result_kind)) and "account" in result_kind.kind_hierarchy():
            account_id_prop = "reported.id"

        async def perform_search(search: str) -> CountByAccount:
            # parse query
            query = await self.template_expander.parse_query(search, on_section="reported", **env)
            # filter only relevant accounts if provided
            if context.accounts:
                query = Query.by(P.single(account_id_prop).is_in(context.accounts)).combine(query)
            # add aggregation to only query for count
            ag_var = AggregateVariable(AggregateVariableName(account_id_prop), "account_id")
            ag_fn = AggregateFunction("sum", 1, [], "count")
            query = evolve(query, aggregate=Aggregate([ag_var], [ag_fn]))
            account_result: CountByAccount = {}
            async with await self.db_access.get_graph_db(graph).search_aggregation(QueryModel(query, model)) as ctx:
                async for result in ctx:
                    account_result[result["group"]["account_id"]] = result["count"] or 0
            return account_result

        async def perform_cmd(cmd: str) -> CountByAccount:
            # filter only relevant accounts if provided
            if context.accounts:
                account_list = ",".join(f'"{a}"' for a in context.accounts)
                cmd = f"search /{account_id_prop} in [{account_list}] | " + cmd
            # aggregate by account
            aggregate = f"aggregate /{account_id_prop} as account_id: sum(1) as count"
            cli_result = await self.cli.execute_cli_command(f"{cmd} | {aggregate}", stream.list, CLIContext(env=env))
            account_result: CountByAccount = {}
            for result in cli_result[0]:
                account_result[result["group"]["account_id"]] = result["count"] or 0
            return account_result

        if resoto_search := inspection.detect.get("resoto"):
            return await perform_search(resoto_search)
        elif resoto_cmd := inspection.detect.get("resoto_cmd"):
            return await perform_cmd(resoto_cmd)
        elif inspection.detect.get("manual"):
            # let's assume the manual check is successful
            return {}
        else:
            raise ValueError(f"Invalid inspection {inspection.id}: no resoto or resoto_cmd defined")

    async def __list_accounts(self, benchmark: Benchmark, graph: GraphName) -> List[str]:
        model = await self.model_handler.load_model(graph)
        gdb = self.db_access.get_graph_db(graph)
        query = Query.by("account")
        if benchmark.clouds:
            query = query.combine(Query.by(P.single("ancestors.cloud.reported.id").is_in(benchmark.clouds)))
        async with await gdb.search_list(QueryModel(query, model)) as crs:
            ids = [value_in_path(a, NodePath.reported_id) async for a in crs]
            return [aid for aid in ids if aid is not None]

    async def validate_benchmark_config(self, json: Json) -> Optional[Json]:
        try:
            benchmark = BenchmarkConfig.from_config(ConfigEntity(ResotoReportBenchmark, json))
            all_checks = {c.id for c in await self.filter_checks()}
            missing = []
            for check in benchmark.nested_checks():
                if check not in all_checks:
                    missing.append(check)
            if missing:
                return {"error": f"Following checks are defined in the benchmark but do not exist: {missing}"}
            else:
                return None
        except Exception as e:
            return {"error": f"Can not digest benchmark: {e}"}

    async def validate_check_collection_config(self, json: Json) -> Optional[Json]:
        try:
            errors = []
            for check in ReportCheckCollectionConfig.from_config(ConfigEntity(ResotoReportCheck, json)):
                try:
                    env = check.default_values or {}
                    if search := check.detect.get("resoto"):
                        await self.template_expander.parse_query(search, on_section="reported", **env)
                    elif cmd := check.detect.get("resoto_cmd"):
                        await self.cli.evaluate_cli_command(cmd, CLIContext(env=env))
                    elif check.detect.get("manual"):
                        pass
                    else:
                        errors.append(f"Check {check.id} neither has a resoto, resoto_cmd or manual defined")
                except Exception as e:
                    errors.append(f"Check {check.id} is invalid: {e}")
            if errors:
                return {"error": f"Can not validate check collection: {errors}"}
            else:
                return None

        except Exception as e:
            return {"error": f"Can not digest check collection: {e}"}



import json
from copy import deepcopy
from datetime import datetime, timedelta
from textwrap import dedent
from typing import Type, Any, Union, cast, List, Dict

import pytest
import yaml
from attr import evolve
from deepdiff import DeepDiff
from hypothesis import HealthCheck, settings, given
from networkx import DiGraph

from resotocore.cli.model import CLIContext
from resotocore.console_renderer import ConsoleRenderer, ConsoleColorSystem
from resotocore.model.graph_access import EdgeTypes
from resotocore.model.model import (
    StringKind,
    Kind,
    NumberKind,
    BooleanKind,
    DateKind,
    DateTimeKind,
    ArrayKind,
    Property,
    ComplexKind,
    Model,
    DictionaryKind,
    predefined_kinds,
    PropertyPath,
    TransformKind,
    DurationKind,
    SyntheticProperty,
    string_kind,
    synthetic_metadata_kinds,
)
from resotocore.model.typed_model import to_json, from_js
from resotocore.types import Json
from resotocore.util import from_utc, utc, utc_str
from tests.resotocore.hypothesis_extension import json_object_gen


def test_json_marshalling() -> None:
    roundtrip(StringKind("string"), Kind)
    roundtrip(StringKind("string", 5, 13, "foo.*bla"), Kind)
    roundtrip(StringKind("string", enum={"foo", "bla"}), Kind)
    roundtrip(NumberKind("num", "int32", minimum=2, maximum=34), Kind)
    roundtrip(NumberKind("num", "int64", enum={1, 2}), Kind)
    roundtrip(BooleanKind("b"), Kind)
    roundtrip(DateKind("d"), Kind)
    roundtrip(DateTimeKind("d"), Kind)
    roundtrip(DurationKind("duration"), Kind)
    roundtrip(TransformKind("synth", "duration", "datetime", "duration_to_datetime", True), Kind)
    roundtrip(ArrayKind(StringKind("string")), Kind)
    roundtrip(Property("foo", "foo"), Property)
    roundtrip(Property("age", "trafo.duration_to_datetime", False, SyntheticProperty(["ctime"])), Property)
    props = [
        Property("array", "string[]"),
        Property("s", "float"),
        Property("i", "int32"),
        Property("other", "SomeComposite"),
    ]
    successor_kinds = {EdgeTypes.default: ["Base", "Test"], EdgeTypes.delete: ["Base"]}
    roundtrip(ComplexKind("Test", ["Base"], props), Kind)
    roundtrip(ComplexKind("Test", [], props, True), Kind)
    roundtrip(ComplexKind("Test", [], props, True, successor_kinds), Kind)


def test_json_sort(person_model: Model) -> None:
    ps = person_model["Person"].sort_json(
        {
            "name": "a",
            "tags": {"c": 2, "a": 1},
            "address": {"name": "a", "id": "a", "city": "gotham", "zip": "123"},
            "mtime": "2020-01-01T00:00:00Z",
            "list": ["b", "a"],
            "id": "a",
        }
    )
    # this is the order of the properties in the model
    assert list(ps.keys()) == ["id", "list", "tags", "mtime", "name", "address"]
    assert list(ps["address"].keys()) == ["id", "zip", "city", "name"]
    # tags do not have a related complex kind: expect natural sort order
    assert list(ps["tags"].keys()) == ["a", "c"]


def test_string() -> None:
    a = StringKind("string", 5, 13, "foo.*bla")
    assert expect_error(a, "foo") == ">foo< does not conform to regex: foo.*bla"
    assert expect_error(a, "fooooo") == ">fooooo< does not conform to regex: foo.*bla"
    assert a.check_valid("fooooobla") is None
    assert expect_error(a, "fooooooooooobla") == ">fooooooooooobla< is too long! Allowed: 13"
    b = StringKind("string", enum={"foo", "bla", "bar"})
    assert b.check_valid("foo") is None
    assert expect_error(b, "baz").startswith(">baz< should be one of")
    assert b.check_valid("$(ENV_VAR)", config_context=True) is None


def test_number() -> None:
    int32 = NumberKind("cores", "int32", 1, 8)
    flot = NumberKind("test", "float", 1, 8)
    assert int32.coerce_if_required(1) is None
    assert int32.coerce_if_required(None) is None
    assert int32.coerce_if_required("no number") is None
    assert int32.check_valid(1) is None
    assert int32.check_valid(8) is None
    assert int32.check_valid("8") is 8
    assert expect_error(int32, "7.123") == "Expected type int32 but got str"
    assert flot.check_valid("7.123") == 7.123
    assert expect_error(int32, 0) == ">0< should be greater or equals than: 1"
    assert expect_error(int32, 9) == ">9< should be smaller or equals than: 8"
    assert expect_error(int32, "9") == ">9< should be smaller or equals than: 8"
    b = NumberKind("bin", "int32", enum={1, 2, 4})
    assert b.check_valid(1) is None
    assert expect_error(b, 3) == ">3< should be one of: {1, 2, 4}"


def test_boolean() -> None:
    a = BooleanKind("question")
    assert a.coerce_if_required(True) is None
    assert a.coerce_if_required(None) is None
    assert a.coerce_if_required("no bool") is None
    assert a.check_valid(True) is None
    assert a.check_valid(False) is None
    assert a.check_valid("true") is True
    assert a.check_valid("false") is False
    assert a.check_valid("FALSE") is False
    assert a.check_valid("$(ENV_VAR)", config_context=True) is None
    assert expect_error(a, "$(ENV_VAR)").startswith("Expected type boolean but got")
    assert expect_error(a, "test").startswith("Expected type boolean but got")


def test_duration() -> None:
    a = DurationKind("dt")
    assert a.check_valid("3d5h6min3s") is None
    assert expect_error(a, True) == "Expected type duration but got bool"
    assert (
        expect_error(a, "23df") == "Wrong format for duration: 23df. Examples: 1yr, 3mo, 3d4h3min1s, 3days and 2hours"
    )
    assert a.coerce("12d") == "1036800s"
    assert a.coerce("12d", normalize=False) == "12d"
    with pytest.raises(AttributeError) as no_date:
        a.check_valid("simply no duration")
    assert (
        str(no_date.value)
        == "Wrong format for duration: simply no duration. Examples: 1yr, 3mo, 3d4h3min1s, 3days and 2hours"
    )


def test_transform() -> None:
    age = TransformKind("dt", "duration", "datetime", "duration_to_datetime", True)
    age.resolve({"duration": DurationKind("duration"), "datetime": DateTimeKind("datetime")})
    with pytest.raises(AttributeError):
        age.check_valid("3s")  # check valid is not allowed on synthetic values (they do not get imported!)
    # age transforms a duration into a timestamp before now
    one_day_old = from_utc(age.coerce_if_required("1d"))  # type: ignore
    # difference between 1d and computed utc-24h should be less than 2 seconds (depending on test env less)
    assert (one_day_old - (utc() - timedelta(hours=24))).total_seconds() <= 2

    # transform back from underlying timestamp to timedelta
    assert age.transform(utc_str(utc() - timedelta(seconds=123))) == "2min3s"
    assert age.transform(utc_str(utc() - timedelta(seconds=123456))) == "1d10h"
    assert age.transform(utc_str(utc() - timedelta(seconds=1234567))) == "14d6h"
    assert age.transform(utc_str(utc() - timedelta(seconds=123456789))) == "3yr10mo"


def test_datetime() -> None:
    a = DateTimeKind("dt")
    assert a.check_valid("2021-06-08T08:56:15Z") is None
    assert a.check_valid("2021-06-08T08:56:15+00:00") == "2021-06-08T08:56:15Z"
    assert expect_error(a, True) == "Expected type datetime but got bool"
    assert a.coerce_if_required(None) is None
    assert a.coerce_if_required("no datetime") is None
    assert a.coerce_if_required("2021-06-08T08") is not None
    assert a.coerce_if_required("2021-06-08T08:56:15Z") is None
    assert a.coerce_if_required("2021-06-08T08:56:15.0000+00:00") == "2021-06-08T08:56:15Z"
    assert a.coerce_if_required("2021-06-08T08:56:15.0000+02:00") == "2021-06-08T06:56:15Z"
    assert a.coerce_if_required("2021-06-08T08:56:15.0000-02:00") == "2021-06-08T10:56:15Z"
    assert a.coerce_if_required("2021-06-08T08:56:15.0000+0000") == "2021-06-08T08:56:15Z"
    assert a.coerce_if_required("2021-06-08 08:56:15").startswith("2021-06-08T")  # type: ignore
    assert a.coerce_if_required("2021-06-08 08:56:15").endswith(":56:15Z")  # type: ignore # ignore the hours, time zone dependant
    today = datetime.today().replace(hour=6, minute=56, second=15).strftime(DateTimeKind.Format)
    assert a.coerce_if_required("08:56:15").startswith(today[0:11])  # type: ignore
    assert a.coerce_if_required("08:56:15").endswith(":56:15Z")  # type: ignore# ignore the hours, time zone dependant
    assert a.coerce_if_required("-12d").startswith("20")  # type: ignore
    assert a.coerce_if_required("12mo").startswith("20")  # type: ignore
    with pytest.raises(Exception) as no_date:
        a.check_valid("simply no date")
    assert str(no_date.value) == f"Invalid isoformat string: 'simply no date'"


def test_date() -> None:
    a = DateKind("d")
    assert a.check_valid("2021-06-08") is None
    assert expect_error(a, True) == "Expected type date but got bool"
    assert a.coerce_if_required("2021-06-08") == "2021-06-08"
    assert a.coerce_if_required("2021 06 08") == "2021-06-08"
    assert a.coerce_if_required("-12d").startswith("20")  # type: ignore
    assert a.coerce_if_required("12mo").startswith("20")  # type: ignore
    with pytest.raises(Exception) as no_date:
        a.check_valid("simply no date")
    assert str(no_date.value) == f"Invalid isoformat string: 'simply no date'"


def test_dictionary() -> None:
    model = {k.fqn: k for k in predefined_kinds}
    result = Property.parse_kind("dictionary[string, string]", model)
    assert isinstance(result, DictionaryKind)
    assert result.key_kind is model["string"]
    assert result.value_kind is model["string"]
    result = Property.parse_kind("dictionary[string, dictionary[string, float]]", model)
    assert isinstance(result, DictionaryKind)
    assert result.key_kind is model["string"]
    assert result.value_kind == DictionaryKind(model["string"], model["float"])
    address = ComplexKind(
        "Foo", [], [Property("tags", "dictionary[string, string]"), Property("anything", "dictionary[string, any]")]
    )
    address_model = Model.from_kinds([address])
    assert address_model.check_valid({"kind": "Foo", "tags": {"a": "b", "b": "c"}}) is None
    assert address_model.check_valid({"kind": "Foo", "tags": {"a": 1, "b": "c"}}) is not None
    assert address_model.check_valid({"kind": "Foo", "anything": {"a": 1, "b": "c", "c": True}}) is None
    expected = 'Kind:Foo Property:anything is not valid: dictionary requires a json object, but got this: 1: {"kind": "Foo", "anything": 1}'
    assert expect_error(address_model, {"kind": "Foo", "anything": 1}) == expected


def test_any() -> None:
    model = Model.from_kinds(predefined_kinds)
    assert model.check_valid({"kind": "any", "a": True, "b": 12, "c": [], "d": {"a": "b"}}) is None


def test_array() -> None:
    foo = ComplexKind("Foo", [], [Property("tags", "dictionary[string, string]"), Property("kind", "string")])
    complex_kind = ComplexKind(
        "TestArray",
        [],
        [
            Property("kind", "string"),
            Property("los", "string[]"),
            Property("lod", "dictionary[string, string][]"),
            Property("foos", "Foo[]"),
            Property("los_los", "string[][]"),
            Property("los_los_los", "string[][][]"),
        ],
    )
    model = Model.from_kinds([foo, complex_kind])
    assert (
        model.check_valid(
            {
                "kind": "TestArray",
                "los": ["a", "b", "c"],
                "lod": [{"a": "b"}, {"b": "c"}],
                "foos": [{"kind": "Foo", "tags": {"a": "b"}}, {"kind": "Foo", "tags": {"b": "c"}}],
                "los_los": [["a", "b"], ["c"], ["d", "e"]],
                "los_los_los": [[["a", "b"], ["c"]], [["d", "e"], ["f"]]],
            }
        )
        is None
    )


def test_model_checking(person_model: Model) -> None:
    assert person_model.check_valid({"kind": "Base", "id": "32"}) is None
    assert person_model.check_valid({"kind": "Base", "id": "32", "list": ["one", "two"]}) is None
    assert person_model.check_valid({"kind": "Base", "id": "32", "list": [1, 2]})["list"] == ["1", "2"]  # type: ignore

    expected = 'Kind:Base Property:list is not valid: Expected property is a json object not an array!: {"kind": "Base", "id": "32", "list": {"not": "an array"}}'
    assert expect_error(person_model, {"kind": "Base", "id": "32", "list": {"not": "an array"}}) == expected
    assert person_model.check_valid({"kind": "Base", "id": 32}) == {"kind": "Base", "id": "32"}
    expected = 'Kind:Base Property:id is required and missing in {"kind": "Base"}'
    assert expect_error(person_model, {"kind": "Base"}) == expected
    expected = "Kind:Base Property:unknown is not defined in model!"
    assert expect_error(person_model, {"kind": "Base", "id": "bla", "unknown": 1}) == expected
    expected = (
        'Kind:Address Property:id is required and missing in {"kind": "Address", "zip": "12345", "city": "gotham"}'
    )
    assert expect_error(person_model, {"kind": "Address", "zip": "12345", "city": "gotham"}) == expected
    nested = {
        "id": "batman",
        "kind": "Person",
        "name": "batman",
        "address": {"kind": "Address", "id": "foo", "city": "gotham"},
    }
    assert person_model.check_valid(nested) is None
    nested = {"id": "batman", "kind": "Person", "name": "batman", "address": {"kind": "Address", "city": "gotham"}}
    expected = 'Kind:Person Property:address is not valid: Kind:Address Property:id is required and missing in {"kind": "Address", "city": "gotham"}: {"id": "batman", "kind": "Person", "name": "batman", "address": {"kind": "Address", "city": "gotham"}}'
    assert expect_error(person_model, nested) == expected
    assert person_model.check_valid({"kind": "Base", "id": "32", "mtime": "2008-09-03T20:56:35+20:00"})["mtime"] == "2008-09-03T00:56:35Z"  # type: ignore
    anything = {"kind": "any", "some": [1, 2, 3], "not": "defined", "props": True}
    assert person_model.check_valid(anything) is None
    any_foo = {"kind": "any_foo", "id": "foo", "foo": {"a": [1, 2, 3]}, "test": "hallo"}
    assert person_model.check_valid(any_foo) is None


def test_property_path() -> None:
    p1 = PropertyPath(["a", None, "c", None])
    p2 = PropertyPath(["a", "b", "c", "d"])
    p3 = PropertyPath(["a", "b"])
    p4 = p3.child("c").child("d")
    assert p1.same_as(p2)
    assert p2.same_as(p1)
    assert not p1.same_as(p3)
    assert p2.same_as(p4)


def test_property_path_on_model(person_model: Model) -> None:
    # complex based property path
    person: ComplexKind = cast(ComplexKind, person_model["Person"])
    person_path = {p.path: p for p in person.resolved_properties()}
    assert len(person_path) == 35
    assert person_path[PropertyPath(["name"])].kind == person_model["string"]
    assert person_path[PropertyPath(["name"])].prop.name == "name"
    assert person_path[PropertyPath(["list[]"])].kind == person_model["string"]
    assert person_path[PropertyPath(["list[]"])].prop.name == "list"
    assert person_path[PropertyPath(["tags", None])].kind == person_model["string"]
    assert person_path[PropertyPath(["address", "zip"])].kind == person_model["zip"]
    assert person_path[PropertyPath(["address", "zip"])].prop.name == "zip"
    with pytest.raises(KeyError):
        _ = person_path[PropertyPath(["anything"])]

    # model based property path
    assert person_model.kind_by_path("name") == person_model["string"]
    assert person_model.kind_by_path("list[]") == person_model["string"]
    assert person_model.kind_by_path("tags.foo") == person_model["string"]
    assert person_model.kind_by_path("tags.bla") == person_model["string"]
    assert person_model.kind_by_path("other_addresses.bla.zip") == person_model["zip"]
    assert person_model.kind_by_path("address.zip") == person_model["zip"]

    # access complex types (user.addresses)
    assert person_model.kind_by_path("addresses") == ArrayKind(person_model["Address"])
    assert person_model.kind_by_path("addresses[23]") == person_model["Address"]
    assert person_model.kind_by_path("addresses[23].zip") == person_model["zip"]
    assert person_model.kind_by_path("other_addresses") == DictionaryKind(string_kind, person_model["Address"])
    assert person_model.kind_by_path("other_addresses.test") == person_model["Address"]

    # properties in hierarchy (tags is defined in base) works as well
    assert person_model.kind_by_path("tags.owner") == person_model["string"]


def test_metadata_on_update(person_model: Model) -> None:
    person = cast(ComplexKind, person_model["Person"])
    person_23 = deepcopy(person)
    person_23.properties = [evolve(prop, metadata={"len": 23}) for prop in person_23.properties]
    person_42 = deepcopy(person)
    person_42.properties = [evolve(prop, metadata={"size": 23, "len": 42}) for prop in person_42.properties]

    def check_person(model: Model, expect_prop: Dict[str, int]) -> None:
        for pp in cast(ComplexKind, model["Person"]).properties:
            assert pp.metadata == expect_prop

    # we update the model and expect the metadata to be set
    updated = person_model.update_kinds([person_23])
    check_person(updated, {"len": 23})
    # when we update the updated model with the original person, we expect the metadata to be unchanged (not removed)
    updated = updated.update_kinds([person])
    check_person(updated, {"len": 23})
    # when we update the updated model with the person_with_bla, we expect the metadata to be updated
    updated = updated.update_kinds([person_42])
    check_person(updated, {"len": 42, "size": 23})
    # when we update the updated model with the person_with_foo, we expect the len property to keep as is
    updated = updated.update_kinds([person_23])
    check_person(updated, {"len": 42})


def test_update(person_model: Model) -> None:
    updated = person_model.update_kinds([StringKind("Foo")])
    assert updated["Foo"].fqn == "Foo"
    # update simple type Foo as Complex is forbidden
    with pytest.raises(AttributeError) as simple:
        updated.update_kinds([ComplexKind("Foo", [], [])])
    assert str(simple.value) == "Update Foo changes an existing property type Foo"
    with pytest.raises(AttributeError) as duplicate:
        updated.update_kinds([ComplexKind("Bla", [], [Property("id", "int32")])])
    assert str(duplicate.value) == "id is defined in graph_root as string and in Bla as int32"

    # update the test property of any_foo from string to an enumeration
    prop = Property("test", "string", description="Some test value.")
    updated.update_kinds(
        [
            StringKind("test_enum", enum={"a", "b", "c"}),
            ComplexKind("any_foo", ["Base"], [evolve(prop, kind="test_enum")]),
        ]
    )
    # allowed to change string to duration, any or date
    for kind in ["duration", "any", "date"]:
        updated.update_kinds([ComplexKind("any_foo", ["Base"], [evolve(prop, kind=kind)])])


def test_load(model_json: str) -> None:
    kinds: List[Kind] = [from_js(a, Kind) for a in json.loads(model_json)]  # type: ignore
    model = Model.from_kinds(kinds)
    assert model.check_valid({"kind": "test.EC2", "id": "e1", "name": "e1", "cores": 1, "mem": 32, "tags": {}}) is None

    base: ComplexKind = model["test.Base"]  # type: ignore
    ec2: ComplexKind = model["test.EC2"]  # type: ignore
    assert ec2.kind_hierarchy() == {"test.Compound", "test.BaseResource", "test.Base", "test.EC2"}
    assert ec2.allow_unknown_props is True
    assert base.allow_unknown_props is False


def test_graph(person_model: Model) -> None:
    graph: DiGraph = person_model.graph()
    assert len(graph.nodes()) == 11
    assert len(graph.edges()) == 9


def test_flatten(person_model: Model) -> None:
    cpl = {m.fqn: m for m in person_model.complex_kinds()}
    flat = {m.fqn: m for m in person_model.flat_kinds() if isinstance(m, ComplexKind)}
    # base property is unchanged
    assert len(flat["Base"].properties) == len(cpl["Base"].properties)
    # address has all properties of base and address
    assert len(flat["Address"].properties) == (len(cpl["Address"].properties) + len(cpl["Base"].properties))
    # address overrides icon in metadata
    assert flat["Address"].metadata["icon"] == "address.svg"
    # person has all properties of base and address
    assert len(flat["Person"].properties) == (len(cpl["Person"].properties) + len(cpl["Base"].properties))
    # person inherits metadata from base
    assert flat["Person"].metadata["icon"] == "icon.svg"


def roundtrip(obj: Any, clazz: Type[object]) -> None:
    js = to_json(obj)
    again = from_js(js, clazz)
    assert type(obj) == type(again)
    assert DeepDiff(obj, again) == {}, f"Json: {js} serialized as {again}"


def expect_error(kind: Union[Kind, Model], obj: Any) -> str:
    try:
        kind.check_valid(obj)
        raise Exception("Expected an error but got a result!")
    except Exception as a:
        return str(a)


@pytest.fixture
def model_json() -> str:
    return """
    [
      {
        "fqn": "test.Compound",
        "properties": [
          { "name": "kind", "kind": "string", "required": true, "description": "The kind of this compound type." }
        ]
      },
      {
        "fqn": "test.Base",
        "properties": [
          { "name": "tags", "kind": "dictionary[string, string]", "description": "Tags that describe the resource.", "required": false }
        ]
      },
      { "fqn" :  "test.BaseResource",
        "bases": ["test.Compound", "test.Base"],
        "properties": [
          { "name": "id", "kind": "string", "description": "The identifier of this resource", "required": true },
          { "name": "name", "kind": "string", "description": "The name of the resource.", "required": true }
        ]
      },
      { "fqn" :  "test.EC2",
        "bases": ["test.BaseResource"],
        "allow_unknown_props": true,
        "properties": [
          { "name": "mem", "kind": "int32", "description": "The amount of bytes", "required": true },
          { "name": "cores", "kind": "int32", "description": "The amount of cores", "required": true }
        ]
      }
    ]
    """


def test_markup() -> None:
    ctx = CLIContext(console_renderer=ConsoleRenderer(color_system=ConsoleColorSystem.monochrome))
    md = dedent(
        """
        - b1 test
        - b2 fox
        - test
        """
    )
    result = ctx.render_console(md)
    assert len(result) > len(md)
    assert "• b1 test" in result


def test_synthetic_predefined(person_model: Model) -> None:
    assert len(person_model.predefined_synthetic_props(synthetic_metadata_kinds)) == len(synthetic_metadata_kinds)


def test_yaml(person_model: Model) -> None:
    person_kind: ComplexKind = person_model["Person"]  # type: ignore
    address = {"zip": "134", "city": "gotham\nmanor house", "number": 123, "float": 1.2345}
    person = {
        "name": "batman",
        "address": address,
        "addresses": [address, address],
        "other_addresses": {"home": address, "work": address},
        "simple": [1, 2, 3, 4, 5, True, False, None],
    }
    assert person_kind.create_yaml(person) == dedent(
        """
        # The name of the person.
        name: 'batman'
        # The address of the person.
        address:
          # The zip code.
          zip: '134'
          # The name of the city.
          # And another line.
          city: |-
            gotham
            manor house

          number: 123
          float: 1.2345
        # The list of addresses.
        addresses:
          - # The zip code.
            zip: '134'
            # The name of the city.
            # And another line.
            city: |-
              gotham
              manor house

            number: 123
            float: 1.2345
          - # The zip code.
            zip: '134'
            # The name of the city.
            # And another line.
            city: |-
              gotham
              manor house

            number: 123
            float: 1.2345
        # Other addresses.
        other_addresses:
          home:
            # The zip code.
            zip: '134'
            # The name of the city.
            # And another line.
            city: |-
              gotham
              manor house

            number: 123
            float: 1.2345
          work:
            # The zip code.
            zip: '134'
            # The name of the city.
            # And another line.
            city: |-
              gotham
              manor house

            number: 123
            float: 1.2345
        simple:
          - 1
          - 2
          - 3
          - 4
          - 5
          - true
          - false
          - null
          """.rstrip()
    )

    assert person == yaml.safe_load(person_kind.create_yaml(person))


@given(json_object_gen)
@settings(max_examples=200, suppress_health_check=list(HealthCheck))
def test_yaml_generation(js: Json) -> None:
    kind = ComplexKind("test", [], [])
    assert js == yaml.safe_load(kind.create_yaml(js))



import re
from datetime import timedelta
from typing import Dict, Optional, List

import pytest

from resotocore.cli.model import CLI
from tests.resotocore.db.entitydb import InMemoryDb
from resotocore.error import NoSuchTemplateError
from resotocore.model.graph_access import Section
from resotocore.query.model import Template, PathRoot
from resotocore.query.template_expander import render_template, TemplateExpanderBase, TemplateExpander
from resotocore.query.template_expander_service import TemplateExpanderService
from resotocore.types import Json
from resotocore.util import utc, from_utc


class InMemoryTemplateExpander(TemplateExpanderBase):
    def __init__(self) -> None:
        self.templates: Dict[str, Template] = {}
        self.props: Json = {}

    async def put_template(self, template: Template) -> None:
        self.templates[template.name] = template

    async def delete_template(self, name: str) -> None:
        self.templates.pop(name, None)

    async def get_template(self, name: str) -> Optional[Template]:
        return self.templates.get(name)

    async def list_templates(self) -> List[Template]:
        return list(self.templates.values())

    def default_props(self) -> Optional[Json]:
        return self.props

    async def parse_query_from_command_line(self, to_parse: str, on_section: str, **env: str) -> str:
        # straight forward implementation: simply strip the "search" command
        return re.sub("^search\\s+", "", to_parse)


@pytest.mark.asyncio
async def test_simple_expand(expander: InMemoryTemplateExpander) -> None:
    templates = [
        Template("foo", "Hey {{name}} - this is {{noun}}"),
        Template("bla", "One, two, {{t3}}"),
        Template("bar", "Heureka"),
    ]
    for t in templates:
        expander.templates[t.name] = t
    result, expands = await expander.expand(
        "Test: expand(foo, name=bart, noun=crazy). expand(bla, t3=jiffy). expand(bar)"
    )
    assert result == "Test: Hey bart - this is crazy. One, two, jiffy. Heureka"
    assert len(expands) == 3
    assert expands[0].template == "foo"
    assert expands[0].props == dict(name="bart", noun="crazy")
    assert expands[1].template == "bla"
    assert expands[1].props == dict(t3="jiffy")
    assert expands[2].template == "bar"
    assert expands[2].props == {}


@pytest.mark.asyncio
async def test_expand(expander: TemplateExpander) -> None:
    await expander.put_template(Template("albs", "is(aws_alb) and age>{{older_than}}"))
    result, expands = await expander.expand("query expand(albs, older_than=7d)")
    assert result == "query is(aws_alb) and age>7d"
    with pytest.raises(NoSuchTemplateError, match="does_not_exist"):
        await expander.expand("query expand(does_not_exist)")


@pytest.mark.asyncio
async def test_add_update_delete_get_list(expander: TemplateExpander) -> None:
    await expander.put_template(Template("albs", "is(aws_alb) and age>{{older_than}}"))
    result = await expander.get_template("albs")
    assert result and result.name == "albs" and result.template == "is(aws_alb) and age>{{older_than}}"
    assert len(await expander.list_templates()) == 1
    await expander.put_template(Template("albs", "is(aws_alb) and age>{{old}} limit 3"))
    assert (await expander.get_template("albs")).template == "is(aws_alb) and age>{{old}} limit 3"  # type: ignore
    await expander.delete_template("albs")
    assert len(await expander.list_templates()) == 0


@pytest.mark.asyncio
async def test_query_creation(cli: CLI) -> None:
    expander = TemplateExpanderService(InMemoryDb(Template, lambda k: k.name), cli)

    async def compare_query(query: str, on_section: str, omit_expansion: bool) -> None:
        direct = await expander.parse_query(query, on_section=on_section, omit_section_expansion=omit_expansion)
        search = await expander.parse_query(
            f"search {query}", on_section=on_section, omit_section_expansion=omit_expansion
        )
        sd = str(direct)
        ss = str(search)[0 : len(sd)]  # strip sort and limit generated by the command line parser
        assert sd == ss

    for section in [*Section.content_ordered, PathRoot]:
        await compare_query('is(aws_alb) and age>3d and mtime>"@UTC@"', section, True)
        await compare_query('is(aws_alb) and age>3d and mtime>"@UTC@"', section, False)


def test_render_simple() -> None:
    attrs = {"foo": "123", "list": ["a", "b", "c"]}
    res = render_template("query foo={{foo}} and test in {{list}}", attrs)
    assert res == 'query foo=123 and test in ["a", "b", "c"]'
    # fallback properties are used if the original list does not contain the value
    res2 = render_template("query foo={{foo}} and test in {{list}}", {}, [attrs])
    assert res2 == 'query foo=123 and test in ["a", "b", "c"]'


def test_render_list() -> None:
    attrs = {"is": ["alb", "elb"]}
    res = render_template("query {{#is.with_index}}{{^first}} or {{/first}}is({{value}}){{/is.with_index}}", attrs)
    assert res == "query is(alb) or is(elb)"


def test_render_as_list() -> None:
    # dictionary as list: every item will be available as [{key: ..., value: ...}]
    attrs: Json = {"props": {"foo": 1, "bla": 2, "bar": 3}}
    res = render_template(
        "query {{#props.as_list.with_index}}{{key}}=={{value}}{{^last}}, {{/last}}{{/props.as_list.with_index}}", attrs
    )
    assert res == "query foo==1, bla==2, bar==3"
    # if the input is already a list, the items will be available as is
    attrs = {"props": [{"key": "foo", "value": 1}, {"key": "bla", "value": 2}, {"key": "bar", "value": 3}]}
    res = render_template(
        "query {{#props.as_list.with_index}}{{key}}=={{value}}{{^last}}, {{/last}}{{/props.as_list.with_index}}", attrs
    )
    assert res == "query foo==1, bla==2, bar==3"
    # if the input is neither dict or list, it will be wrapped in a single element list
    attrs = {"props": "test"}
    res = render_template("query {{#props.as_list.with_index}}{{value}}=={{index}}{{/props.as_list.with_index}}", attrs)
    assert res == "query test==0"


def test_from_now() -> None:
    res = render_template("{{delta.from_now}}", {"delta": "4h"})
    in_4_hours = utc() + timedelta(hours=4)
    assert abs((in_4_hours - from_utc(res)).total_seconds()) < 1


def test_render_filter() -> None:
    attrs = {"foo": "123", "filter": 32}
    template = "query foo={{foo.parens}}{{#filter}} and some.other.prop == {{filter}}{{/filter}}"
    res = render_template(template, attrs)
    assert res == 'query foo="123" and some.other.prop == 32'
    attrs2 = {"foo": "123"}
    res = render_template(template, attrs2)
    assert res == 'query foo="123"'


def test_custom_tags() -> None:
    res = render_template("@test@ and @rest@", dict(test="work", rest="play"), tags=("@", "@"))
    assert res == "work and play"


@pytest.mark.asyncio
async def test_abbreviation(expander: TemplateExpander) -> None:
    async def with_section(section: Optional[str]) -> None:
        # name form
        assert str(await expander.parse_query("cloud=a", section)) == 'ancestors.cloud.reported.name == "a"'
        assert str(await expander.parse_query("account=a", section)) == 'ancestors.account.reported.name == "a"'
        assert str(await expander.parse_query("region=a", section)) == 'ancestors.region.reported.name == "a"'
        assert str(await expander.parse_query("zone=a", section)) == 'ancestors.zone.reported.name == "a"'
        # short form
        assert str(await expander.parse_query("cloud.name=a", section)) == 'ancestors.cloud.reported.name == "a"'
        assert str(await expander.parse_query("account.name=a", section)) == 'ancestors.account.reported.name == "a"'
        assert str(await expander.parse_query("region.name=a", section)) == 'ancestors.region.reported.name == "a"'
        assert str(await expander.parse_query("zone.name=a", section)) == 'ancestors.zone.reported.name == "a"'
        assert str(await expander.parse_query("usage.cpu.max > 3", section)) == "usage.cpu.max > 3"

    for s in Section.content_ordered + [None]:
        await with_section(s)



from resotocore.model.resolve_in_graph import GraphResolver


def test_resolved_ancestors() -> None:
    assert GraphResolver.resolved_ancestors == {
        "account": "refs.account_id",
        "cloud": "refs.cloud_id",
        "region": "refs.region_id",
        "zone": "refs.zone_id",
    }



from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EntityDb, EventEntityDb, ArangoEntityDb
from resotocore.ids import SubscriberId
from resotocore.task.model import Subscriber

SubscriberDb = EntityDb[SubscriberId, Subscriber]
EventSubscriberDb = EventEntityDb[SubscriberId, Subscriber]


def subscriber_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[SubscriberId, Subscriber]:
    return ArangoEntityDb(db, collection, Subscriber, lambda k: k.id)



import asyncio
from typing import List

from jsons import DeserializationError
from pytest import mark, raises

from resotocore.analytics import AnalyticsEvent, CoreEvent
from resotocore.config import ConfigHandler, ConfigEntity
from resotocore.config.core_config_handler import CoreConfigHandler
from resotocore.core_config import (
    ResotoCoreConfigId,
    ResotoCoreRoot,
    ResotoCoreCommandsRoot,
    CustomCommandsConfig,
    ResotoCoreSnapshotsRoot,
)
from resotocore.system_start import empty_config
from resotocore.message_bus import MessageBus, CoreMessage
from resotocore.ids import ConfigId


@mark.asyncio
async def test_model_updated_on_start(core_config_handler: CoreConfigHandler, config_handler: ConfigHandler) -> None:
    try:
        default_model = await config_handler.get_configs_model()
        await core_config_handler.start()
        assert len(await config_handler.get_configs_model()) > len(default_model)
    finally:
        await core_config_handler.stop()


@mark.asyncio
async def test_config_ingested_on_start(core_config_handler: CoreConfigHandler, config_handler: ConfigHandler) -> None:
    try:
        assert await config_handler.get_config(ResotoCoreConfigId) is None
        await core_config_handler.start()
        assert await config_handler.get_config(ResotoCoreConfigId) is not None
    finally:
        await core_config_handler.stop()


@mark.asyncio
async def test_exit_on_updated_config(
    core_config_handler: CoreConfigHandler, message_bus: MessageBus, core_config_handler_exits: List[bool]
) -> None:
    callback_result = []

    async def callback(config_id: ConfigId) -> None:
        callback_result.append(config_id)

    try:
        await core_config_handler.start()
        core_config_handler.add_callback(callback)
        await asyncio.sleep(0.1)
        await message_bus.emit_event(CoreMessage.ConfigUpdated, {"id": ResotoCoreConfigId})
        await asyncio.sleep(0.1)
        assert len(core_config_handler_exits) == 1
        assert len(callback_result) == 1
        assert callback_result[0] == ResotoCoreConfigId
    finally:
        await core_config_handler.stop()


@mark.asyncio
async def test_validation(core_config_handler: CoreConfigHandler) -> None:
    validate = core_config_handler.validate_config_entry

    # empty config is valid
    assert await validate({"config": {ResotoCoreRoot: {}}}) is None
    # expected json object but got 23
    assert await validate({"config": {ResotoCoreRoot: 23}}) is not None
    # validation fails, since ui-path does not exist
    assert await validate({"config": {ResotoCoreRoot: {"api": {"tsdb_proxy_url": "wrong"}}}}) is not None
    # default configuration is valid
    assert await validate({"config": {ResotoCoreRoot: empty_config().json()}}) is None

    # empty command config is fine
    assert await validate({"config": {ResotoCoreCommandsRoot: {}}}) is None
    # 23 can not be parsed as command list
    with raises(DeserializationError):
        await validate({"config": {ResotoCoreCommandsRoot: {"commands": 23}}})
    # valid entry can be read
    assert await validate({"config": {ResotoCoreCommandsRoot: CustomCommandsConfig().json()}}) is None

    # cron expression is valid
    assert (
        await validate({"config": {ResotoCoreSnapshotsRoot: {"foo-label": {"schedule": "0 0 * * *", "retain": 42}}}})
        is None
    )

    # cron expression is invalid
    assert (
        await validate({"config": {ResotoCoreSnapshotsRoot: {"foo-label": {"schedule": "foo bar", "retain": 42}}}})
        is not None
    )


@mark.asyncio
async def test_detect_usage_metrics_turned_off(
    config_handler: ConfigHandler, core_config_handler_started: CoreConfigHandler
) -> None:
    # make sure usage metrics are enabled
    core_config_handler_started.config.runtime.usage_metrics = True
    await config_handler.patch_config(
        ConfigEntity(ResotoCoreConfigId, {ResotoCoreRoot: {"runtime": {"usage_metrics": True}}})
    )
    # disable usage metrics
    await config_handler.patch_config(
        ConfigEntity(ResotoCoreConfigId, {ResotoCoreRoot: {"runtime": {"usage_metrics": False}}})
    )
    await asyncio.sleep(0.1)
    events: List[AnalyticsEvent] = core_config_handler_started.event_sender.events  # type: ignore # in-memory sender
    assert len(events) == 1
    assert events[0].kind == CoreEvent.UsageMetricsTurnedOff



from typing import Union, List, Any, AsyncIterator
from resotocore.infra_apps.manifest import AppManifest
from resotocore.types import JsonElement, Json
from resotocore.ids import GraphName
from attrs import frozen
from abc import ABC, abstractmethod


@frozen
class Failure:
    error: str


@frozen
class Success:
    output: List[Any]


AppResult = Union[Failure, Success]


class Runtime(ABC):
    """
    Runtime is the interface to run an infrastructure app.
    """

    @abstractmethod
    async def execute(
        self,
        graph: GraphName,
        manifest: AppManifest,
        config: Json,
        stdin: AsyncIterator[JsonElement],
        argv: List[str],
        ctx: Any,  # CLIContext, but here we use Any to avoid circular dependency
    ) -> AsyncIterator[JsonElement]:
        """
        Executes the infrastructure app."""
        yield None

    @abstractmethod
    async def generate_template(
        self,
        graph: GraphName,
        manifest: AppManifest,
        config: Json,
        stdin: AsyncIterator[JsonElement],
        argv: List[str],
    ) -> AsyncIterator[str]:
        """
        Generates the template for the infrastructure app. Does not execute any commands
        """
        yield ""



from __future__ import annotations

import calendar
import inspect
import json
from abc import ABC, abstractmethod
from asyncio import iscoroutine
from datetime import timedelta
from enum import Enum
from functools import reduce
from pathlib import Path
from textwrap import dedent
from typing import (
    Optional,
    List,
    Any,
    Dict,
    Tuple,
    Callable,
    Union,
    Awaitable,
    Type,
    cast,
    Set,
    AsyncIterator,
    TYPE_CHECKING,
)

from aiostream import stream
from aiostream.core import Stream
from attrs import define, field
from parsy import test_char, string
from rich.jupyter import JupyterMixin

from resotocore.cli import JsGen, T, Sink
from resotocore.console_renderer import ConsoleRenderer, ConsoleColorSystem
from resotocore.core_config import AliasTemplateConfig, AliasTemplateParameterConfig
from resotocore.error import CLIParseError
from resotocore.ids import GraphName
from resotocore.query.model import Query, variable_to_absolute, PathRoot
from resotocore.query.template_expander import render_template
from resotocore.types import Json, JsonElement
from resotocore.util import AccessJson, uuid_str, from_utc, utc, utc_str
from resotolib.parse_util import l_curly_dp, r_curly_dp
from resotolib.utils import get_local_tzinfo

if TYPE_CHECKING:
    from resotocore.dependencies import Dependencies


class MediaType(Enum):
    Json = 1
    FilePath = 2
    Markdown = 3
    String = 4

    @property
    def text(self) -> bool:
        return self in (MediaType.Json, MediaType.String, MediaType.Markdown)

    @property
    def file_path(self) -> bool:
        return self == MediaType.FilePath

    def __repr__(self) -> str:
        return "application/json" if self == MediaType.Json else "application/octet-stream"


no_closing_p = test_char(lambda x: x != "}", "No closing bracket").at_least(1).concat()
no_bracket_p = test_char(lambda x: x not in ("{", "}"), "No opening bracket").at_least(1).concat()
double_curly_open_dp = string("{{")
double_curly_close_dp = string("}}")
l_or_r_curly_dp = string("{") | string("}")

# use this property name as reference to self when defined via .
self_name = "self_" + uuid_str()


@define(frozen=True)
class FilePath:
    user: Path
    local: Path

    def json(self) -> Json:
        return {"user_path": str(self.user), "local_path": str(self.local)}

    @staticmethod
    def from_path(in_path: JsonElement) -> FilePath:
        if isinstance(in_path, str):
            p = Path(in_path)
            return FilePath(Path(p.name), p.expanduser().absolute())
        elif isinstance(in_path, dict) and "user_path" in in_path and "local_path" in in_path:
            return FilePath(Path(in_path["user_path"]), Path(in_path["local_path"]).expanduser().absolute())
        else:
            raise ValueError(f"Invalid file path: {in_path}")

    @staticmethod
    def user_local(user: Union[str, Path], local: Union[str, Path]) -> FilePath:
        return FilePath(Path(user), Path(local).expanduser().absolute())


@define(frozen=True)
class CLIContext:
    env: Dict[str, str] = field(factory=dict)
    uploaded_files: Dict[str, str] = field(factory=dict)  # id -> path
    query: Optional[Query] = None
    query_options: Dict[str, Any] = field(factory=dict)
    commands: List[ExecutableCommand] = field(factory=list)
    console_renderer: Optional[ConsoleRenderer] = None
    source: Optional[str] = None  # who is calling

    @property
    def graph_name(self) -> GraphName:
        return GraphName(self.env["graph"])

    def variable_in_section(self, variable: str) -> str:
        # if there is no entity provider, always assume the root section
        section = (
            self.env.get("section")
            if self.query or self.commands and isinstance(self.commands[0].command, EntityProvider)
            else PathRoot
        )
        return variable_to_absolute(section, variable)

    def render_console(self, element: Union[str, JupyterMixin]) -> str:
        if self.console_renderer:
            return self.console_renderer.render(element)
        elif isinstance(element, JupyterMixin):
            return str(element)
        else:
            return element

    def text_generator(
        self, line: ParsedCommandLine, in_stream: AsyncIterator[JsonElement]
    ) -> AsyncIterator[JsonElement]:
        async def render_markdown() -> AsyncIterator[str]:
            async for e in in_stream:
                yield self.render_console(e)  # type: ignore

        if line.produces == MediaType.Markdown:
            return render_markdown()
        else:
            return in_stream

    def supports_color(self) -> bool:
        return (
            self.console_renderer is not None
            and self.console_renderer.color_system is not None
            and self.console_renderer.color_system != ConsoleColorSystem.monochrome
        )

    def formatter(self, format_string: str) -> Callable[[Json], str]:
        return self.formatter_with_variables(format_string, False)[0]

    def formatter_with_variables(
        self, format_string: str, collect_variables: bool = True
    ) -> Tuple[Callable[[Json], str], Optional[Set[str]]]:
        """
        A renderer can be used to string format objects based on a provided format string.
        """

        variables: Optional[Set[str]] = set() if collect_variables else None

        def format_variable(name: str) -> str:
            assert "__" not in name, "No dunder attributes allowed"
            if name in (".", "/"):
                in_section = self_name
            else:
                in_section = self.variable_in_section(name)
                if collect_variables:
                    variables.add(in_section)  # type: ignore
            return "{" + in_section + "}"

        def render_simple_property(prop: Any) -> str:
            return json.dumps(prop) if prop is None or isinstance(prop, bool) else str(prop)

        variable = (l_curly_dp >> no_closing_p << r_curly_dp).map(format_variable)
        token = double_curly_open_dp | double_curly_close_dp | no_bracket_p | variable | l_or_r_curly_dp
        format_string_parser = token.many().concat()
        formatter: str = format_string_parser.parse(format_string)

        def format_object(obj: Any) -> str:
            return formatter.format_map(AccessJson.wrap(obj, "null", render_simple_property, self_name))

        return format_object, variables


EmptyContext = CLIContext()


class CLIEngine(ABC):
    @abstractmethod
    async def evaluate_cli_command(
        self, cli_input: str, context: CLIContext = EmptyContext, replace_place_holder: bool = True
    ) -> List[ParsedCommandLine]:
        pass


@define
class CLICommandRequirement:
    name: str


@define
class CLIFileRequirement(CLICommandRequirement):
    path: str  # local client path


class CLIAction(ABC):
    def __init__(
        self, produces: MediaType, requires: Optional[List[CLICommandRequirement]], envelope: Optional[Dict[str, str]]
    ) -> None:
        self.produces = produces
        self.required = requires or []
        self.envelope: Dict[str, str] = envelope or {}

    @staticmethod
    def make_stream(in_stream: JsGen) -> Stream:
        return in_stream if isinstance(in_stream, Stream) else stream.iterate(in_stream)


class CLISource(CLIAction):
    def __init__(
        self,
        fn: Callable[[], Union[Tuple[Optional[int], JsGen], Awaitable[Tuple[Optional[int], JsGen]]]],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(produces, requires, envelope)
        self._fn = fn

    async def source(self) -> Tuple[Optional[int], Stream]:
        res = self._fn()
        count, gen = await res if iscoroutine(res) else res
        return count, self.make_stream(await gen if iscoroutine(gen) else gen)

    @staticmethod
    def no_count(
        fn: Callable[[], Union[JsGen, Awaitable[JsGen]]],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> CLISource:
        return CLISource.with_count(fn, None, produces, requires, envelope)

    @staticmethod
    def with_count(
        fn: Callable[[], Union[JsGen, Awaitable[JsGen]]],
        count: Optional[int],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> CLISource:
        async def combine() -> Tuple[Optional[int], JsGen]:
            res = fn()
            gen = await res if iscoroutine(res) else res
            return count, gen

        return CLISource(combine, produces, requires, envelope)

    @staticmethod
    def single(
        fn: Callable[[], Union[JsGen, Awaitable[JsGen]]],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> CLISource:
        return CLISource.with_count(fn, 1, produces, requires, envelope)

    @staticmethod
    def empty() -> CLISource:
        return CLISource.with_count(stream.empty, 0)


class CLIFlow(CLIAction):
    def __init__(
        self,
        fn: Callable[[JsGen], Union[JsGen, Awaitable[JsGen]]],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(produces, requires, envelope)
        self._fn = fn

    async def flow(self, in_stream: JsGen) -> Stream:
        gen = self._fn(self.make_stream(in_stream))
        return self.make_stream(await gen if iscoroutine(gen) else gen)


@define
class ArgInfo:
    # If the argument has a name. It is quite common that arguments do not have a name
    # but are expected at some position.
    # Example: `count <kind>`: kind is the argument at position 1 without a name
    name: Optional[str] = None
    # Defines if this argument expects a value.
    # Some arguments are only flags, while others expect a value.
    # Example: `--compress` is a flag without value
    # Example: `--count <kind>` is an argument with value
    expects_value: bool = False
    # If the value has to be picked from a list of values (enumeration).
    # Example: `--format svg|png|jpg`
    possible_values: List[str] = field(factory=list)
    # If this argument is allowed to be specified multiple times
    can_occur_multiple_times: bool = False
    # Give a type hint for the argument value.
    # Allowed values are:
    # - `file`: the argument expects a file path
    # - `kind`: the argument expects a kind in the model
    # - `property`: the argument expects a property in the model
    # - `command`: the argument expects a command on the cli
    # - `event`: the event handled or emitted by the task handler
    # - `search`: the argument expects a search string
    value_hint: Optional[str] = None
    # Help text of the argument option.
    help_text: Optional[str] = None
    # If multiple options share the same group, only one of them can be selected.
    # Use groups if you have multiple options, where only one is allowed to be selected.s
    option_group: Optional[str] = None


# mypy does not support recursive type aliases: define 3 levels as maximum here
ArgsInfo = Union[
    Dict[str, Union[Dict[str, Union[Dict[str, Union[Any, List[ArgInfo]]], List[ArgInfo]]], List[ArgInfo]]],
    List[ArgInfo],
]


class CLICommand(ABC):
    """
    The CLIPart is the base for all participants of the cli execution.
    Source: generates a stream of objects
    Flow: transforms the elements in a stream of objects
    Sink: takes a stream of objects and creates a result
    """

    def __init__(
        self, dependencies: Dependencies, category: str = "misc", allowed_in_source_position: bool = False
    ) -> None:
        self.dependencies = dependencies
        self.category = category
        self.allowed_in_source_position = allowed_in_source_position

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    def help(self) -> str:
        # if not defined in subclass, fallback to inline doc
        doc = inspect.getdoc(type(self))
        return doc if doc else f"{self.name}: no help available."

    def rendered_help(self, ctx: CLIContext) -> str:
        text = f"\n**{self.name}: {self.info()}**\n\n{self.help()}"
        return ctx.render_console(text)

    @abstractmethod
    def args_info(self) -> ArgsInfo:
        pass

    @abstractmethod
    def info(self) -> str:
        pass

    @abstractmethod
    def parse(self, arg: Optional[str] = None, ctx: CLIContext = EmptyContext, **kwargs: Any) -> CLIAction:
        pass

    @staticmethod
    def get_from(name: str, kind: Type[T], kwargs: Dict[str, Any]) -> Optional[T]:
        if (kd := kwargs.get(name)) and isinstance(kd, kind):
            return kd
        return None

    @staticmethod
    def get_previous_command(kwargs: Dict[str, Any]) -> Optional[ExecutableCommand]:
        return CLICommand.get_from("previous_command", ExecutableCommand, kwargs)


@define(order=True, hash=True, frozen=True)
class AliasTemplateParameter:
    name: str
    description: str
    default: Optional[JsonElement] = None

    def example_value(self) -> JsonElement:
        return self.default if self.default else f"test-{self.name.replace('_', '-')}"

    @property
    def arg_name(self) -> str:
        return "--" + self.name.replace("_", "-")


@define(order=True, hash=True, frozen=True)
class InfraAppAliasParameter:
    name: str
    help: str
    default: Optional[JsonElement]

    @property
    def arg_name(self) -> str:
        return "--" + self.name.replace("_", "-")


# pylint: disable=not-an-iterable
@define(order=True, hash=True, frozen=True)
class InfraAppAlias:
    name: str
    description: str
    readme: str
    parameters: List[InfraAppAliasParameter]

    def template(self) -> str:
        return f"apps run {self.name}" + r" {{args}}"

    def render(self, props: Json) -> str:
        return render_template(self.template(), props)

    def help(self) -> str:
        args = " ".join(f"{arg.arg_name} <value>" for arg in self.parameters)

        indent = "        "
        desc = ""
        if self.readme:
            for line in self.readme.splitlines():
                desc += f"\n{indent}{line}"

        def param_info_infra_apps(param: InfraAppAliasParameter) -> str:
            default = f" [default: {param.default}]" if param.default else " [required]"
            return f"- `{param.arg_name}`{default}: {param.help}"

        arg_info = f"\n{indent}".join(param_info_infra_apps(param) for param in (self.parameters or []))
        result = dedent(
            f"""
        {self.name}: {self.description}
        ```shell
        {self.name} {args}
        ```
        {desc}
        ## Parameters
        {arg_info}"""
        )
        return result

    def rendered_help(self, ctx: CLIContext) -> str:
        return ctx.render_console(self.help())


# pylint: disable=not-an-iterable
@define(order=True, hash=True, frozen=True)
class AliasTemplate:
    name: str
    info: str
    template: str
    parameters: List[AliasTemplateParameter] = field(factory=list)
    description: Optional[str] = None
    # only use args_description if the template does not use explicit parameters
    args_description: Dict[str, str] = field(factory=dict)
    allowed_in_source_position: bool = False

    def render(self, props: Json) -> str:
        return render_template(self.template, props)

    def args_info(self) -> ArgsInfo:
        args_desc = [ArgInfo(name, expects_value=True, help_text=desc) for name, desc in self.args_description.items()]
        param = [
            ArgInfo(
                p.arg_name,
                expects_value=True,
                help_text=f"[{'required' if p.default is None else 'optional'}] {p.description}",
            )
            for p in sorted(self.parameters, key=lambda p: p.default is not None)  # required parameters first
        ]
        return args_desc + param

    def help_with_params(self) -> str:
        args = " ".join(f"{arg.arg_name} <value>" for arg in self.parameters)

        def param_info(p: AliasTemplateParameter) -> str:
            default = f" [default: {p.default}]" if p.default else " [required]"
            return f"- `{p.arg_name}`{default}: {p.description}"

        def sort_required_name(p: AliasTemplateParameter) -> Any:
            return p.default is not None, p.name

        indent = "            "
        arg_info = f"\n{indent}".join(param_info(arg) for arg in sorted(self.parameters, key=sort_required_name))
        minimal = " ".join(f'{p.arg_name} "{p.example_value()}"' for p in self.parameters if p.default is None)
        desc = ""
        if self.description:
            for line in self.description.splitlines():
                desc += f"\n{indent}{line}"

        return dedent(
            f"""
            {self.name}: {self.info}
            ```shell
            {self.name} {args}
            ```
            {desc}
            ## Parameters
            {arg_info}

            ## Template
            ```shell
            > {self.template}
            ```

            ## Example
            ```shell
            # Executing this command
            > {self.name} {minimal}
            # Will expand to this command
            > {self.render({p.name: p.example_value() for p in self.parameters})}
            ```
            """
        )

    def help_no_params_args(self) -> str:
        args = ""
        args_info = ""
        for arg_name, arg_description in self.args_description.items():
            args += f" [{arg_name}]"
            args_info += f"\n- `{arg_name}`: {arg_description}"

        args_info = args_info or ("<args>" if "{args}" in self.template else "")
        return (
            f"{self.name}: {self.info}\n```shell\n{self.name} {args}\n```\n\n"
            f"## Parameters\n{args_info}\n\n{self.description}\n\n"
        )

    def help(self) -> str:
        return self.help_with_params() if self.parameters else self.help_no_params_args()

    def rendered_help(self, ctx: CLIContext) -> str:
        return ctx.render_console(self.help())

    @staticmethod
    def from_config(cfg: AliasTemplateConfig) -> AliasTemplate:
        def arg(p: AliasTemplateParameterConfig) -> AliasTemplateParameter:
            return AliasTemplateParameter(p.name, p.description, p.default)

        return AliasTemplate(
            name=cfg.name,
            info=cfg.info,
            template=cfg.template,
            parameters=[arg(a) for a in cfg.parameters],
            description=cfg.description,
            allowed_in_source_position=cfg.allowed_in_source_position or False,
        )


@define
class WorkerCustomCommand:
    """
    A worker might provide custom commands. This definition is provided by the worker.
    """

    name: str
    info: Optional[str] = None
    args_description: Dict[str, str] = field(factory=dict)
    description: Optional[str] = None
    filter: Dict[str, List[str]] = field(factory=dict)
    allowed_on_kind: Optional[str] = None
    expect_node_result: bool = False

    def to_template(self) -> AliasTemplate:
        allowed_kind = f" --allowed-on {self.allowed_on_kind}" if self.allowed_on_kind else ""
        result_flag = "" if self.expect_node_result else " --no-node-result"
        command = f"--command '{self.name}'"
        args = "--arg '{{args}}'"
        return AliasTemplate(
            name=self.name,
            info=self.info or "",
            args_description=self.args_description,
            template=f"execute-task{result_flag}{allowed_kind} {command} {args}",
            description=self.description,
        )


class InternalPart(ABC):
    """
    Internal parts can be executed but are not shown via help.
    They usually get injected by the CLI Interpreter to ease usability.
    """


class EntityProvider(ABC):
    """
    Mark this command as a provider of entities with: id, reported, desired, metadata.
    """


class OutputTransformer(ABC):
    """
    Mark all commands that transform the output stream (formatting).
    """


class PreserveOutputFormat(ABC):
    """
    Mark all commands where the output should not be flattened to default line output.
    """


class NoTerminalOutput(ABC):
    """
    Mark all commands where the output should not contain any terminal escape codes.
    """


@define
class ParsedCommand:
    cmd: str
    args: Optional[str] = None


@define
class ParsedCommands:
    commands: List[ParsedCommand]
    env: Json = field(factory=dict)


@define
class ExecutableCommand:
    name: str  # the name of the command or alias
    command: CLICommand
    arg: Optional[str]
    action: CLIAction


@define
class ParsedCommandLine:
    """
    The parsed command line holds:
    - ctx: the resulting environment coming from the parsed environment + the provided environment
    - commands: all commands this command is defined from
    - generator: this generator can be used in order to execute the command line
    """

    ctx: CLIContext
    parsed_commands: ParsedCommands
    executable_commands: List[ExecutableCommand]
    unmet_requirements: List[CLICommandRequirement]
    envelope: Dict[str, str]

    def __attrs_post_init__(self) -> None:
        def expect_action(cmd: ExecutableCommand, expected: Type[T]) -> T:
            action = cmd.action
            if isinstance(action, expected):
                return action
            else:
                message = "must be the first command" if issubclass(type(action), CLISource) else "no source data given"
                raise CLIParseError(f"Command >{cmd.command.name}< can not be used in this position: {message}")

        if self.executable_commands:
            expect_action(self.executable_commands[0], CLISource)
            for command in self.executable_commands[1:]:
                expect_action(command, CLIFlow)

    async def to_sink(self, sink: Sink[T]) -> T:
        _, generator = await self.execute()
        return await sink(generator)

    @property
    def commands(self) -> List[CLICommand]:
        return [part.command for part in self.executable_commands]

    @property
    def produces(self) -> MediaType:
        # the last command in the chain defines the resulting media type
        return self.executable_commands[-1].action.produces if self.executable_commands else MediaType.Json

    async def execute(self) -> Tuple[Optional[int], Stream]:
        if self.executable_commands:
            source_action = cast(CLISource, self.executable_commands[0].action)
            count, flow = await source_action.source()
            for command in self.executable_commands[1:]:
                flow_action = cast(CLIFlow, command.action)
                flow = await flow_action.flow(flow)
            return count, flow
        else:
            return 0, stream.empty()


class CLI(ABC):
    @abstractmethod
    async def evaluate_cli_command(
        self, cli_input: str, context: CLIContext = EmptyContext, replace_place_holder: bool = True
    ) -> List[ParsedCommandLine]:
        pass

    @abstractmethod
    async def execute_cli_command(self, cli_input: str, sink: Sink[T], ctx: CLIContext = EmptyContext) -> List[T]:
        pass

    @abstractmethod
    def register_infra_app_alias(self, alias: InfraAppAlias) -> None:
        pass

    @abstractmethod
    def unregister_infra_app_alias(self, name: str) -> None:
        pass

    @abstractmethod
    def register_alias_template(self, template: AliasTemplate) -> None:
        """
        Called when something introduces a custom command.
        """

    @abstractmethod
    def unregister_alias_template(self, name: str) -> None:
        """
        Called when something removes a custom command.
        """

    @property
    @abstractmethod
    def direct_commands(self) -> Dict[str, CLICommand]:
        pass

    @property
    @abstractmethod
    def alias_commands(self) -> Dict[str, CLICommand]:
        pass

    @property
    @abstractmethod
    def commands(self) -> Dict[str, CLICommand]:
        pass

    @property
    @abstractmethod
    def env(self) -> Dict[str, Any]:
        pass

    @property
    @abstractmethod
    def dependencies(self) -> Any:  # CliDependencies, but we can't import it here
        pass

    @property
    @abstractmethod
    def alias_templates(self) -> Dict[str, AliasTemplate]:
        pass

    @property
    @abstractmethod
    def infra_app_aliases(self) -> Dict[str, InfraAppAlias]:
        pass

    @staticmethod
    def replacements(**env: str) -> Dict[str, str]:
        now_string = env.get("now")
        ut = from_utc(now_string) if now_string else utc()
        t = ut.date()
        try:
            n = ut.astimezone(get_local_tzinfo())
        except Exception:
            n = ut
        return dict(
            UTC=utc_str(ut),
            NOW=n.strftime("%Y-%m-%dT%H:%M:%S%z"),
            TODAY=t.strftime("%Y-%m-%d"),
            TOMORROW=(t + timedelta(days=1)).isoformat(),
            YESTERDAY=(t + timedelta(days=-1)).isoformat(),
            YEAR=t.strftime("%Y"),
            MONTH=t.strftime("%m"),
            DAY=t.strftime("%d"),
            TIME=n.strftime("%H:%M:%S"),
            HOUR=n.strftime("%H"),
            MINUTE=n.strftime("%M"),
            SECOND=n.strftime("%S"),
            TZ_OFFSET=n.strftime("%z"),
            TZ=n.strftime("%Z"),
            MONDAY=(t + timedelta((calendar.MONDAY - t.weekday()) % 7)).isoformat(),
            TUESDAY=(t + timedelta((calendar.TUESDAY - t.weekday()) % 7)).isoformat(),
            WEDNESDAY=(t + timedelta((calendar.WEDNESDAY - t.weekday()) % 7)).isoformat(),
            THURSDAY=(t + timedelta((calendar.THURSDAY - t.weekday()) % 7)).isoformat(),
            FRIDAY=(t + timedelta((calendar.FRIDAY - t.weekday()) % 7)).isoformat(),
            SATURDAY=(t + timedelta((calendar.SATURDAY - t.weekday()) % 7)).isoformat(),
            SUNDAY=(t + timedelta((calendar.SUNDAY - t.weekday()) % 7)).isoformat(),
        )

    @staticmethod
    def replace_placeholder(cli_input: str, **env: str) -> str:
        # We do not use the template renderer here on purpose:
        # - the string is processed before it is evaluated - there is no way to escape the @ symbol
        # - the string might contain @ symbols
        result = reduce(lambda res, kv: res.replace(f"@{kv[0]}@", kv[1]), CLI.replacements(**env).items(), cli_input)
        result = reduce(
            lambda res, kv: res.replace(f"@{kv[0].lower()}@", kv[1]), CLI.replacements(**env).items(), result
        )
        return result



from typing import List, AsyncIterator, Type, Optional, Any
from resotocore.infra_apps.runtime import Runtime
from resotocore.infra_apps.manifest import AppManifest
from resotocore.service import Service
from resotocore.types import Json, JsonElement
from resotocore.ids import GraphName
from resotocore.db.model import QueryModel
from resotocore.cli.model import CLI, CLIContext
from resotocore.cli import NoExitArgumentParser
from jinja2 import Environment
import logging
from aiostream.core import Stream
from aiostream import stream
from argparse import Namespace
from resotolib.durations import parse_optional_duration
from resotolib.asynchronous.utils import async_lines
from pydoc import locate


log = logging.getLogger(__name__)


class LocalResotocoreAppRuntime(Runtime, Service):
    """
    Runtime implementation that runs the Infrastructure Apps directly on the resotocore.
    Currently, only the Jinja2 language is supported.
    """

    def __init__(self, cli: CLI) -> None:
        self.cli = cli
        self.dbaccess = cli.dependencies.db_access
        self.model_handler = cli.dependencies.model_handler
        self.template_expander = cli.dependencies.template_expander

    async def execute(
        self,
        graph: GraphName,
        manifest: AppManifest,
        config: Json,
        stdin: AsyncIterator[JsonElement],
        argv: List[str],
        ctx: CLIContext,
    ) -> AsyncIterator[JsonElement]:
        """
        Runtime implementation that runs the app locally.
        """
        async for line in self.generate_template(graph, manifest, config, stdin, argv):
            async with (await self._interpret_line(line, ctx)).stream() as streamer:
                async for item in streamer:
                    yield item

    async def generate_template(
        self,
        graph: GraphName,
        manifest: AppManifest,
        config: Json,
        stdin: AsyncIterator[JsonElement],
        argv: List[str],
    ) -> AsyncIterator[str]:
        graphdb = self.dbaccess.get_graph_db(graph)
        env = Environment(extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"], enable_async=True)
        template = env.from_string(manifest.source)

        model = await self.model_handler.load_model(graph)

        async def perform_search(search: str) -> AsyncIterator[Json]:
            # parse query
            query = await self.template_expander.parse_query(search, on_section="reported")
            async with await graphdb.search_graph_gen(QueryModel(query, model)) as ctx:
                async for result in ctx:
                    yield result

        template.globals["parse_duration"] = parse_optional_duration
        template.globals["search"] = perform_search

        args = self._args_from_manifest(manifest, argv)

        async for line in async_lines(template.generate_async(config=config, args=args, stdin=stdin)):
            line = line.strip()
            log.debug(f"Rendered infrastructure app line: {line}")
            if not line:
                continue
            yield line

    def _args_from_manifest(self, manifest: AppManifest, argv: Optional[List[str]] = None) -> Namespace:
        args_schema = manifest.args_schema or {}

        parser = NoExitArgumentParser(description=manifest.description)

        def str_to_type(type_str: Optional[str]) -> Optional[Type[Any]]:
            if type_str is None:
                return None
            supported_types = {"bool", "str", "int", "float", "complex"}
            if type_str not in supported_types:
                raise ValueError(f"Unsupported type: {type_str}")
            return locate(type_str)  # type: ignore

        for arg_name, arg_info in args_schema.items():
            kwargs = {}
            for flag in ["help", "action", "default", "type", "nargs", "required"]:
                if flag in arg_info:
                    if flag == "type":
                        kwargs[flag] = str_to_type(arg_info[flag])
                    else:
                        kwargs[flag] = arg_info[flag]
            parser.add_argument(f"--{arg_name}", **kwargs)  # type: ignore

        return parser.parse_args(argv)

    async def _interpret_line(self, line: str, ctx: CLIContext) -> Stream:
        command_streams: List[Stream] = []
        total_nr_outputs: int = 0
        parsed_commands = await self.cli.evaluate_cli_command(line, ctx, True)
        for parsed in parsed_commands:
            nr_outputs, command_output_stream = await parsed.execute()
            total_nr_outputs = total_nr_outputs + (nr_outputs or 0)
            command_streams.append(command_output_stream)

        return stream.concat(stream.iterate(command_streams), task_limit=1)
